#include "barrier.h"
#include <stdint.h>
#include <iostream>
#include <pthread.h>

my_barrier *b;
int counter_first;
int counter_second;
int failures;
int iteration_count;
pthread_mutex_t failure_lock = PTHREAD_MUTEX_INITIALIZER;

void *thread_loop(void *raw_thread_id) {
    int thread_id = reinterpret_cast<intptr_t>(raw_thread_id);
    for (int i = 0; i < iteration_count; ++i) {
        if (counter_first != i) {
            pthread_mutex_lock(&failure_lock);
            std::cerr << "thread " << thread_id << ": counter_first=" << counter_first
                      << "; versus i=" << i << std::endl;
            ++failures;
            pthread_mutex_unlock(&failure_lock);
        }
        if (thread_id == 0) {
            ++counter_second;
        }
        barrier_wait(b);
        if (counter_second - 1 != i) {
            pthread_mutex_lock(&failure_lock);
            std::cerr << "thread " << thread_id << ": counter_second=" << counter_second
                      << "but should be " << i - 1 << std::endl;
            ++failures;
            pthread_mutex_unlock(&failure_lock);
        }
        if (thread_id == 0) {
            ++counter_first;
        }
        barrier_wait(b);
    }
    return 0;
}

bool test_loop(int num_threads, int num_iterations) {
    counter_first = counter_second = failures = 0;
    iteration_count = num_iterations;
    pthread_t *threads = new pthread_t[num_threads];
    std::cout << "testing " << num_threads << " threads and " << (num_iterations * 2) << " waits ..." << std::flush;
    b = barrier_create(num_threads);
    for (int i = 0; i < num_threads; ++i) {
        pthread_create(&threads[i], 0, thread_loop, (void*) (intptr_t) i);
    }
    for (int i = 0; i < num_threads; ++i) {
        pthread_join(threads[i], 0);
    }
    barrier_destroy(b);
    if (failures == 0) {
        std::cout << "passed" << std::endl;
        return true;
    } else {
        std::cout << "failed" << std::endl;
        return false;
    }
}

void *thread_one_barrier(void *raw_thread_id) {
    int thread_id = reinterpret_cast<intptr_t>(raw_thread_id);
    if (counter_first != 0) {
        pthread_mutex_lock(&failure_lock);
        std::cerr << "counter_first=" << counter_first << " but should be 0 (should only change when all threads reach barrier)\n";
        ++failures;
        pthread_mutex_unlock(&failure_lock);
    }
    barrier_wait(b);
    if (thread_id == 0) {
        ++counter_first;
    }
    return 0;
}

bool test_once(int num_threads, int attempt) {
    failures = counter_first = 0;
    pthread_t *threads = new pthread_t[num_threads];
    std::cout << "testing " << num_threads << " threads and using barrier once (attempt " << attempt 
              << ")..." << std::flush;
    b = barrier_create(num_threads);
    for (int i = 0; i < num_threads; ++i) {
        pthread_create(&threads[i], 0, thread_one_barrier, (void*) (intptr_t) i);
    }
    for (int i = 0; i < num_threads; ++i) {
        pthread_join(threads[i], 0);
        ++counter_first;
    }
    barrier_destroy(b);
    if (failures == 0) {
        std::cout << "passed" << std::endl;
        return true;
    } else {
        std::cout << "failed" << std::endl;
        return false;
    }
}

int main(int argc, char **argv) {
    int failed = 0;
    int passed = 0;
    int skipped = 0;

    /* test using barrier once from 1 to 4 threads, 4 times */
    for (int j = 0; j < 4; ++j) {
        for (int i = 1; i <= 4; ++i) {
            if (!test_once(i, j)) {
                ++failed;
            } else {
                ++passed;
            }
        }
    }
    
    /* test using barrier twice from 1 to 4 threads 
       (only if using it once worked) */
    if (failed == 0) {
        for (int i = 1; i <= 4; ++i) {
            if (!test_loop(i, 1)) {
                ++failed;
            } else {
                ++passed;
            }
        }
    } else {
        skipped += 4;
    }

    /* test using barrier four times from 1 to 4 threads 
       (only if using it twice worked) */
    if (failed == 0) {
        for (int i = 1; i <= 4; ++i) {
            if (!test_loop(i, 2)) {
                ++failed;
            } else {
                ++passed;
            }
        }
    } else {
        skipped += 4;
    }

    /* test using barrier 20k times from 1 to 4 threads,
       and 200 times from 32 threads 
       (only if prior tests passed */
    if (failed == 0) {
        for (int i = 1; i <= 4; ++i) {
            if (!test_loop(i, 10000)) {
                ++failed;
            } else {
                ++passed;
            }
        }
        if (!test_loop(32, 100)) {
            ++failed;
        } else {
            ++passed;
        }
    } else {
        skipped += 5;
    }

    if (!failed) {
        std::cout << "*** all tests passed\n";
    } else {
        std::cout << "*** " << passed << " tests passed; " << failed << " tests failed; " << skipped << " tests skipped\n";
    }
    return 0;
}
